#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fitsio.h"
#include "saotemp.h"

#define LINEPIX 1024
#define LINEMODEL 11500
#define MAXCHAR 1024
#define MAXHDR 1024
#define HDRSIZE 80
/*#define LIBPATH "/home/motohara/lib/saotemp" */

main(int argc, char **argv)
{
  fitsfile *fp;
  FILE *fpmodel;
  int i,j;
  int flag;
  int nmodel;
  double ipixmin,ipixmax;
  double lam,lammin,lammax,pixsum;
  double fac;
  int status,nfound,anynull,bitpix=-32;
  long fpixel, nelements,naxis=2,npixels;
  long naxes[2]={LINEPIX, LINEPIX},npix;
  float nullval; 

  float *pix,*spix;
  double *mod_lam, *mod_flux;
  double *flux;
  double sflux;
  char line[MAXCHAR];

  int islit;
  double offset,slit,teff;
  double a,b,c;

  char openfile[100],savefile[100],modelfile[100];
  char grism[100];
  char filter02[64];

  int nkeys;
  char *hdrptr[MAXHDR];

  mod_lam=(double *) malloc(LINEMODEL*sizeof(double));
  mod_flux=(double *) malloc(LINEMODEL*sizeof(double));

  if(argc!=7 && argc!=8){
    fprintf(stderr,"Usage : %s input_file output_file stellar_type slit_width(pix) offset factor_of_bb [zJGr/JHGr/wKGr]\n",argv[0]);
    exit(1);
  }

  sprintf(openfile, "%s",argv[1]);
  sprintf(savefile, "%s",argv[2]);
  slit=atof(argv[4]);
  offset=atof(argv[5]);
  fac=atof(argv[6]);
  if(argc==8) sprintf(grism, "%s",argv[7]);
  
  if(strcmp(argv[3],"A0")==0) teff=9480;
  else if(strcmp(argv[3],"A2")==0) teff=8810;
  else if(strcmp(argv[3],"A5")==0) teff=8160;
  else if(strcmp(argv[3],"A7")==0) teff=7930;
  else if(strcmp(argv[3],"F0")==0) teff=7020;
  else if(strcmp(argv[3],"F2")==0) teff=6750;
  else if(strcmp(argv[3],"F5")==0) teff=6530;
  else if(strcmp(argv[3],"F7")==0) teff=6240;
  else if(strcmp(argv[3],"G0V")==0) teff=5930;
  else if(strcmp(argv[3],"G2V")==0) teff=5830;
  else if(strcmp(argv[3],"G4V")==0) teff=5740;
  else if(strcmp(argv[3],"G6V")==0) teff=5620;
  else if(strcmp(argv[3],"G0III")==0) teff=5910;
  else if(strcmp(argv[3],"G4III")==0) teff=5190;
  else if(strcmp(argv[3],"G6III")==0) teff=5050;
  else if(strcmp(argv[3],"G8III")==0) teff=4960;
  else {
    fprintf(stderr,"type must be A0/A2/A5/A7/F0/F2/F5/F7/G0V/G2V/G4V/G6V/G0III/G4III/G6III/G8III\n");
    exit(1);
  }

  /* load model spectrum */
  sprintf(modelfile,"%s/model_%s.dat",LIBPATH,argv[3]);
  fprintf(stderr,"%s \t: Read %s  ... ",argv[0],modelfile);
  if((fpmodel=fopen(modelfile,"r"))==NULL){
    fprintf(stderr,"Open Error  %s",modelfile);
    exit(1);
  }
  i=0;
  nmodel=0;
  while(fgets(line,MAXCHAR,fpmodel)!=NULL){
    if(i>2){
      sscanf(line,"%lf %lf",mod_lam,mod_flux);
      *mod_lam/=10000.0;
      mod_lam++;
      mod_flux++;
      nmodel++;
    }
    i++;
  }
  mod_lam-=nmodel;
  mod_flux-=nmodel;

  fclose(fpmodel);
  fprintf(stderr,"OK\n");
  /* load spectrum data */
  fprintf(stderr,"%s \t: Read %s  ... \n",argv[0],openfile);
  status=0;
  if(fits_open_file(&fp,openfile, READONLY, &status)){
    fits_report_error(stderr,status);
    return(status);
  }

  if(argc==7){
    if(fits_read_key_str(fp, "FILTER02" , grism, NULL, &status)){
      fits_report_error(stderr,status);
      return(status);
    }
    fprintf(stderr,"### According to the header, grism is %s ###\n",grism);
  }


  /* define lamda-pix relation */
  /* also add 0.2 to fac when the grism is zJGr */
  /*  if(strcmp(grism,"zJGr")==0){a=3.58446141e-8; b=5.16578664e-4; c=8.44173704e-1;} */
  if(strcmp(grism,"zJGr")==0){
    a=3.828795e-08; b=5.133416e-04; c=8.447779e-01;
    fac+=1.0;
  } 
  else if(strcmp(grism,"JHGr")==0){
    a=3.644766e-08; b=8.165282e-04; c=1.017306e+00;
  }
  else if(strcmp(grism,"wKGr")==0){
    a=6.927999e-08; b=8.249877e-04; c=1.703723e+00;
  }

  else{
    fprintf(stderr,"Error : Grism should be zJGr, JHGr or wKGr.\n");
    exit(1);
  }

  if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  if((naxes[0]!=1)&&(naxes[1]!=1)){
    fprintf(stderr,"1-D image only.\n");
    exit(1);
  }
  
  if(fits_get_hdrspace(fp, &nkeys, NULL, &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  for (i=0;i<nkeys;i++){
    hdrptr[i]=malloc(HDRSIZE);
    if(fits_read_record(fp,i+1,hdrptr[i],&status)){
      fits_report_error(stderr,status);
      return(status);
    }
  }

  npix=naxes[0]*naxes[1];
  pix=(float *) malloc(npix*sizeof(float));
  flux=(double *) malloc(npix*sizeof(double));
  fpixel = 1;
  nullval = 0;
  if (fits_read_img(fp, TFLOAT, fpixel, npix, &nullval, pix, &anynull, &status)){
    fits_report_error(stderr,status);
    return(status);
  }

  if(fits_close_file(fp, &status)){
    fits_report_error(stderr,status);
    return(status);
  }

  /* calculate model flux for each pixel*/
  for(i=0;i<npix;i++){
    ipixmin=i+1-0.5;
    ipixmax=i+1+0.5;
    lammin=ipixmin*ipixmin*a+ipixmin*b+c;
    lammax=ipixmax*ipixmax*a+ipixmax*b+c;
    lam=((double) i)*((double) i)*a+((double) i)*b+c;
    /*    fprintf(stdout,"%d %e\n",i,(lammax+lammin)/2); */
    flux[i]=0;
    flag=0;
    for(j=0;j<nmodel;j++){
      if(lammin<=mod_lam[j] && mod_lam[j]<=lammax){
	if(flag==0){
	  flux[i]=(mod_flux[j]+((mod_flux[j]-mod_flux[j-1])*lammin-mod_flux[j]*mod_lam[j-1]+mod_flux[j-1]*mod_lam[j])/(mod_lam[j]-mod_lam[j-1]))*(mod_lam[j]-lammin); 
	  flag=1;
	}
	else flux[i]+=(mod_flux[j]+mod_flux[j-1])*(mod_lam[j]-mod_lam[j-1]);
      }
      else if (flag==1 && lammax<mod_lam[j]){
	flux[i]+=(mod_flux[j-1]+((mod_flux[j]-mod_flux[j-1])*lammax-mod_flux[j]*mod_lam[j-1]+mod_flux[j-1]*mod_lam[j])/(mod_lam[j]-mod_lam[j-1]))*(lammax-mod_lam[j-1]);
	break;
      }
    }
    flux[i]=flux[i]/(lammax-lammin);
    /*    fprintf(stdout,"%e %e\n",flux[i],6.1e15*fac/pow(lam,5)/(exp(14387.7/lam/teff)-1.0)); */
    flux[i]+=6.1e15*fac/pow(lam,5)/(exp(14387.7/lam/teff)-1.0);
    
  }
  
  /*smooth model flux to slit width and modify pixel value*/
  islit=floor(slit/2.0);
  pixsum=0;
  for(i=islit;i<npix-islit;i++){
    sflux=0;
    for(j=i-islit;j<=i+islit;j++) sflux+=flux[j]; 
    pix[i]=pix[i]/sflux;
    pixsum+=pix[i];
  }
  for(i=0;i<islit;i++) pix[i]=pix[i]/flux[i];
  for(i=npix-islit;i<npix;i++) pix[i]=pix[i]/flux[i];


  /* normalization */
  for(i=0;i<npix;i++){
    pix[i]=pix[i]/(pixsum / ((double) (npix+1)));
  }

  fprintf(stderr,"\t: Write %s  ... ",savefile);
  unlink(savefile);

  status=0;
  if (fits_create_file(&fp, savefile, &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  for (i=0;i<nkeys;i++){
    if(fits_write_record(fp,hdrptr[i],&status)){
      fits_report_error(stderr,status);
      return(status);
    }
  }
  fpixel=1;
  npixels=naxes[0]*naxes[1];
  fprintf(stderr,"npix %d\n",npixels);
  if (fits_write_img(fp, TFLOAT, fpixel, npixels, pix, &status)) 
    fits_report_error(stderr,status);

  if (fits_close_file(fp, &status)){
    fits_report_error(stderr,status);
    return(status);
  }

  for (i=0;i<nkeys;i++) free(hdrptr[i]);
  free(mod_lam);
  free(mod_flux);
  free(flux);
  free(pix);

}
